
#include <stdbool.h>	/* bool */
#include <stdio.h>		/* perror, fprintf */
#include <stdlib.h>		/* malloc */
#include <string.h>		/* strlen, strncpy */

#define HASH

#ifdef HASH

#include <search.h>

#else /* ! def HASH */
#error "regex is obsolete, use HASH"

#include <sys/types.h>
#include <regex.h>

#define regcheck(ecode, desc, ...) \
		do { \
			if (!ecode) break; \
			char ebuf[80]; \
			regerror(ecode, &preg, ebuf, sizeof(ebuf)); \
			fprintf(stderr, desc ": %s\n", ##__VA_ARGS__, ebuf); \
			exit(EXIT_FAILURE); \
		} while (0)
#endif /* ! def HASH */


#define NEXTERN
#include "formparse.h"
#undef NEXTERN


#ifdef DEBUG

#define dbg0(fmt, ...) \
		fprintf(stdout, "%s:%u\t- symbol: %2d [%u:%u] -- " fmt "\n", \
				__FUNCTION__, __LINE__, tok.type, tok.from, tok.to, ##__VA_ARGS__)
#define dbg() dbg0("\b\b\b   ")

#else /* ! def DEBUG */

#define dbg0(fmt, ...)
#define dbg()

#endif /* ! def DEBUG */


/* ------------------------------------------------------------------------------------- */
/*
	expr ::= ["+"|"-"] term1 [("+"|"-") expr] .
	term1 ::= term2 {("*"|"/") term2}
	term2 ::= factor ["^" term2)] .
	factor ::= factor0 ["!"]
	factor0 ::= ident | number | numberf | "(" expr ")" | func "(" expr ")" .
*/

typedef enum {
	unknown, end_of_input,
	IDENT, NUMBER, NUMBERF, LPARAN, RPARAN, FUNC,
	EMARK, HAT, TIMES, SLASH, PLUS, MINUS
} symbol;
static const char *symbol_names[] = {
	"unknown", "end of input",
	"identifier", "number", "float-number", "(", ")", "function",
	"!", "^", "*", "/", "+", "-"
};

typedef struct {
	symbol type;
	unsigned int from, to;
	union {
		unsigned int ival;
		float fval;
	} v;
} token;


static void scanner();
static void deinit();
static void init(char const *, int, char const **, int, char const **);
static inline ptree_t * tree_node(fsym, unsigned int, ptree_t *);
#ifndef HASH
static void construct_regex(int, char const **, int, char const **);
#endif
static bool match(symbol);
static ptree_t * expr(ptree_t *);
static ptree_t * factor0(ptree_t *);
static ptree_t * factor(ptree_t *);
static ptree_t * term1(ptree_t *);
static ptree_t * term2(ptree_t *);

#define expect(sym) (match(sym) || ( \
			fprintf(stderr, "expected " #sym " at %d of '%s' (in %s:%d)\n", \
					tok.from, org, __FUNCTION__, __LINE__), \
			false))

static token tok;									// the current token
static char const *c, *org;							// pointer to the formula to parse

#ifdef HASH
#define min(a, b) ((a < b) ? a : b)
typedef union {
	void *ptr;
	struct {
		bool isfunc : 1;
		unsigned int idx : min(sizeof(int),(sizeof(void *)*8-1));
	} data;
} tokentry_t;
#undef min
static int fac0len_max;
static char *buf;
#else
static char *regbuf;
static regex_t preg;
static int fncs;
static int vars;
static regmatch_t *pmatch;
#endif

/* ------- */
/* SCANNER */
/* ------- */

static void scanner() {
	tok.from = c - org;
	tok.type = unknown;
	tok.v.ival = 0;
	
	char fc = *c++;
	if (('A' <= fc && fc <= 'Z') || ('a' <= fc && fc <= 'z')) {
#ifdef HASH
		ENTRY e, *ep;
		tokentry_t te;
		char const *start;
		
		start = c - 1;
		while (('A' <= *c && *c <= 'Z') || ('a' <= *c && *c <= 'z') || ('0' <= *c && *c <= '9'))
			c++;
		if (c - start > fac0len_max) {
			fprintf(stderr, "cannot identify factor '%s'\n", start);
			exit(EXIT_FAILURE);
		}
		
		strncpy(buf, start, c - start);
		buf[c - start] = '\0';
		e.key = buf;
		ep = hsearch(e, FIND);
		if (!ep) {
			fprintf(stderr, "%s '%s' at character %d of input not recognized\n",
					(*c == '(') ? "function" : "identifier", buf, start - org + 1);
			exit(EXIT_FAILURE);
		}
		te.ptr = ep->data;
		tok.type = (te.data.isfunc) ? FUNC : IDENT;
		tok.v.ival = te.data.idx;
#else
		// may be IDENT or FUNC or garbage
		int ecode;
		int nmatch = fncs + 2;
		// index 0 holds the complete match, 1 the group of variables and 2-fncs the group of functions
		
		dbg0(stderr, "checking '%s' with regex...\n", c - 1);
		ecode = regexec(&preg, c - 1, nmatch, pmatch, 0);
		regcheck(ecode, "error matching sub-expression '%s'; expected IDENT or FUNC", c - 1);
		
		if (pmatch[1].rm_so != -1) {
			dbg0(stderr, "found variable %c\n", c[-1]);
			tok.type = IDENT;
			tok.v.ival = c[-1] - 'x';			// store index of variable, 0: x, 1: y, 2: z
			c += pmatch[1].rm_eo - 1;			// always 0, since our variables x, y and z have length 1
		} else {
			int i;
			bool found = false;
			for (i=0; i<fncs; i++) {
				if (pmatch[2 + i].rm_so != -1) {
					dbg0(stderr, "found func #%d\n", i);
					found = true;
					tok.type = FUNC;
					tok.v.ival = i;
					c += pmatch[2 + i].rm_eo - 1;
					break;
				}
			}
			if (!found)
				fprintf(stderr, "function referenced at '%s' not recognized, this might be an internal error!\n", c - 1);
		}
#endif
	} else if (('0' <= fc && fc <= '9') || fc == '.') {
		// NUMBER and NUMBERF
		char *endptr;
		tok.type = NUMBER;
		tok.v.ival = strtoul(c - 1, &endptr, 10);
		if (*endptr == '.') {
			char *endptr2;
			float f = strtof(c - 1, &endptr2);
			if (endptr2 > endptr + 1) {				// don't treat values like '7.' as NUMBERF
				tok.type = NUMBERF;
				tok.v.fval = f;
			}
			endptr = endptr2;
		}
		c = endptr;
		
	} else {
		// operators / parantheses
		switch (fc) {
			case '(': tok.type = LPARAN; break;
			case ')': tok.type = RPARAN; break;
			case '!': tok.type = EMARK; break;
			case '^': tok.type = HAT; break;
			case '*': tok.type = TIMES; break;
			case '/': tok.type = SLASH; break;
			case '+': tok.type = PLUS; break;
			case '-': tok.type = MINUS; break;
			
			case '\0': tok.type = end_of_input; break;
			
			default:
				// error
				fprintf(stderr, "unrecognized character '%c' at %d, remaining: '%s'\n", c[-1], tok.from + 1, c - 1);
				exit(EXIT_FAILURE);
		}
	}
	
	tok.to = c - org;
	dbg0("'%s'", org + tok.from);
}

/* ------ */
/* PARSER */
/* ------ */

static bool match(symbol tt) {
	if (tok.type != tt) {
		// error
		return false;
	}
	scanner();
	return true;
}


static ptree_t * factor0(ptree_t *par) {	
	dbg();
	
	ptree_t *r = NULL;
	int val = tok.v.ival;
	
	if (match(IDENT)) {
		r = tree_node(VAR, val, par);
	
	} else if (match(NUMBER)) {
		r = tree_node(NUM, val, par);
	
	} else if (match(NUMBERF)) {
		r = tree_node(NUMF, val, par);
	
	} else if (match(LPARAN)) {
		r = expr(par);
		if (!expect(RPARAN)) return NULL;
	
	} else if (match(FUNC)) {
		r = tree_node(FUN, val, par);
		if (!expect(LPARAN)) return NULL;
		r->right = expr(r);
		if (!expect(RPARAN)) return NULL;
	
	} else {
		// error
		if (tok.type == end_of_input) {
			fprintf(stderr, "premature end of formula, expected factor0\n");
			return r;
		}
		fprintf(stderr, "ignoring illegal symbol '%s' at %d of input '%s'\n", symbol_names[tok.type], tok.from + 1, org);
		scanner();
		return factor0(par);
	}
	return r;
}

static ptree_t * factor(ptree_t *par) {
	dbg();
	
	ptree_t *r = factor0(par);
	if (match(EMARK)) {
		ptree_t *v = tree_node(FAC, 0, par);
		v->left = r;
		r->up = v;
		r = v;
	}
	return r;
}

static ptree_t * term2(ptree_t *par) {
	dbg();
	
	ptree_t *r = factor(par);
	if (tok.type == HAT) {
		ptree_t *v = tree_node(EXP, 0, par);
		v->left = r;
		r->up = v;
		scanner();
		v->right = term2(v);
		r = v;
	}
	return r;
}

// todo: transform like the one below

static ptree_t * term1(ptree_t *par) {
	dbg();
	
	ptree_t *r = term2(par);
	while (tok.type == TIMES || tok.type == SLASH) {
		ptree_t *v = tree_node((tok.type == TIMES) ? MUL : DIV, 0, par);
		v->left = r;
		r->up = v;
		scanner();
		v->right = term2(v);
		r = v;
	}
	return r;
}

static ptree_t * expr(ptree_t *par) {
	dbg();
	
	ptree_t *r = NULL;
	
	if (tok.type == PLUS)
		scanner();
	if (tok.type == MINUS) {
		ptree_t *v = tree_node(SUB, 0, par);
		v->left = tree_node(NUM, 0, v);
		
		scanner();
		r = v;
	}
	
	if (r == NULL) {
		r = term1(par);
	} else {
		r->right = term1(r);
	}
	
	if (tok.type == PLUS || tok.type == MINUS) {
		ptree_t *v = tree_node(ADD, 0, par);
		v->left = r;
		r->up = v;
		v->right = expr(v);
		r = v;
	}
	
	return r;
}

/* ------- */
/* HELPERS */
/* ------- */

static inline ptree_t * tree_node(fsym type, unsigned int value, ptree_t *par) {
	ptree_t *t = malloc(sizeof(ptree_t));
	if (t == NULL) {
		perror("malloc");
		exit(EXIT_FAILURE);
	}
	t->left = NULL;
	t->right = NULL;
	t->up = par;
	t->sym = type;
	t->v.ival = value;
	return t;
}

#ifndef HASH
static void construct_regex(int vrsc, char const **vsrv, int fncc, char const **fncv) {
	static const char regvar[] = "^\\(x\\|y\\|z\\)";
	
	char *rb;
	size_t regfun_len;
	int i;
	int *fnlens;
	
	fnlens = malloc(sizeof(int) * fncc);
	regfun_len = 0;									// "\|\("
	for (i=0; i<fncc; i++) {
		fnlens[i] = strlen(fncv[i]);
		regfun_len += fnlens[i] + 6;			// "\|\(" and "\)"
	}
	
	// the space for the trailing '\0' is provided by sizeof(regvar)
	regbuf = malloc(sizeof(regvar) + regfun_len);
	
	// copy the initial regex for variables x, y and z
	strncpy(regbuf, regvar, sizeof(regvar) - 1);
	rb = regbuf + sizeof(regvar) - 1;
	
	// copy the function names, each in an own group
	for (i=0; i<fncc; i++) {
		*rb++ = '\\';
		*rb++ = '|';
		*rb++ = '\\';
		*rb++ = '(';
		// size_t len = strlen(fncv[i]);
		strncpy(rb, fncv[i], fnlens[i]);
		rb += fnlens[i];
		*rb++ = '\\';
		*rb++ = ')';
	}
	
	*rb = '\0';
	
	dbg0("constructed regex '%s'", regbuf);
	free(fnlens);
	
	i = regcomp(&preg, regbuf, 0);
	regcheck(i, "error compiling regex '%s'", regbuf);
}
#endif

static void init(char const *f, int vrsc, char const **vrsv, int fncc, char const **fncv) {
	c = f;
	org = f;
	
#ifdef HASH
	hcreate((vrsc + fncc) * 2);
	ENTRY e;
	tokentry_t te;
	int i, n;
	fac0len_max = 0;
	te.data.isfunc = false;
	for (i=0; i<vrsc; i++) {
		te.data.idx = i;
		if ((n = strlen(vrsv[i])) > fac0len_max) fac0len_max = n;
		e.key = (char *)vrsv[i];
		e.data = te.ptr;
		hsearch(e, ENTER);
	}
	te.data.isfunc = true;
	for (i=0; i<fncc; i++) {
		te.data.idx = i;
		if ((n = strlen(fncv[i])) > fac0len_max) fac0len_max = n;
		e.key = (char *)fncv[i];
		e.data = te.ptr;
		hsearch(e, ENTER);
	}
	buf = malloc(fac0len_max + 1);
#else
	// +2 to accommodate group 0 (whole match) and group 1 (variables)
	pmatch = malloc(sizeof(regmatch_t) * (fncc + 2));
	fncs = fncc;
	construct_regex(vrsc, vrsv, fncc, fncv);
#endif
	
	scanner();
}

static void deinit() {
#ifdef HASH
	hdestroy();
	free(buf);
#else
	regfree(&preg);
	free(pmatch);
	free(regbuf);
#endif
}

/* ------- */
/* EXPORTS */
/* ------- */

ptree_t * parse(const char *s, int nvars, char const **vars, int nfuncs, char const **funcs) {
	init(s, nvars, vars, nfuncs, funcs);
	ptree_t *r = expr(NULL);
	if (tok.type != end_of_input)
		fprintf(stderr,
				"cannot parse garbage at the end of the input: '%s', ignoring...\n",
				c - 1);
	deinit();
	return r;
}

const char * const fsym_chrs = "vnNf!^*/+-";
int ptree_output_paran = 0;

void ptree_output(ptree_t *t) {
	int v = t->v.ival;
	switch (t->sym) {
		case VAR: printf("[%d]", v); break;
		case NUM: printf((v >= 10) ? "(%d)" : "%d", v); break;
		case NUMF: printf("(%g)", t->v.fval); break;
		case FUN: printf("{%d}", v); break;
		
		case FAC: case EXP: case MUL:
		case DIV: case ADD: case SUB:
			printf("%c", fsym_chrs[t->sym]);
	}
}



void ptree_walk(ptree_t *t, tnode_vis pre, tnode_vis in, tnode_vis post) {
	if (t == NULL) return;
	if (t->left == NULL && t->right == NULL) {
		if (pre) pre(t);
		if (in) in(t);
		if (post) post(t);
	} else {
		if (t->left != NULL && t->left->up != t) {
			printf("internal error: t = ");
			ptree_output(t);
			printf(", left child = ");
			ptree_output(t->left);
			printf(" doesn't have t as parent, but ");
			if (t->left->up == NULL) {
				printf("NULL");
			} else {
				ptree_output(t->left->up);
			}
			printf("\n");
			exit(EXIT_FAILURE);
		}
		
		if (t->right != NULL && t->right->up != t) {
			printf("internal error: t = ");
			ptree_output(t);
			printf(", right child = ");
			ptree_output(t->right);
			printf(" doesn't have t as parent, but ");
			if (t->right->up == NULL) {
				printf("NULL");
			} else {
				ptree_output(t->right->up);
			}
			printf("\n");
			exit(EXIT_FAILURE);
		}
		
		if (t->up && ptree_output_paran) printf("(");
		if (pre) pre(t);
		ptree_walk(t->left, pre, in, post);
		if (in) in(t);
		ptree_walk(t->right, pre, in, post);
		if (post) post(t);
		if (t->up && ptree_output_paran) printf(")");
	}
}

#ifdef BUILD_FPMAIN
int main(int argc, char const **argv) {
	if (argc != 2)
		exit(1);
	
	char const *funcs[] = { "sin", "cos", "bla", "blubb", "blo", "och" };
	int nfuncs = sizeof(funcs) / sizeof(funcs[0]);
	char const *vars[] = { "x", "y", "z", "e", "pi", "i" };
	int nvars = sizeof(vars) / sizeof(vars[0]);
	
	ptree_t *t = parse(argv[1], nvars, vars, nfuncs, funcs);
	
	tnode_vis in = NULL, post = NULL;
	if (true) {
		ptree_output_paran = true;
		in = ptree_output;
	} else {
		post = ptree_output;
	}
	ptree_walk(t, NULL, in, post);
	printf("\n");
	
	ptree_output_paran = false;
	ptree_walk(t, NULL, NULL, (tnode_vis)free);
}
#endif
