
#include <stdbool.h>
#include <time.h>

#include <SDL/SDL.h>
#include <SDL/SDL_rotozoom.h>

#include "newton.h"

#define CAPTION "Newton fractals"
#define TITLE   "Newton fractals"

#define min(a, b) (((a) < (b)) ? (a) : (b))
#define max(a, b) (((a) > (b)) ? (a) : (b))

#define sdl_fail(func) do { \
			fprintf(stderr, "%s failed: %s\n", func, SDL_GetError()); \
			exit(EXIT_FAILURE); } while (0)
#define sdl_run(failres, func, ...) do { \
			if (SDL_ ## func(__VA_ARGS__) == failres) \
				sdl_fail("SDL_" #func); } while (0)


static void mainloop(SDL_Surface * const, int const, render_t *);
static SDL_Surface * sdl_init(int, int, int);

static SDL_Surface *display = NULL;

#define HSV_ACC 16384
static void setpixel(SDL_Surface *img, SDL_Color *c, unsigned int x, unsigned int y) {
	if (img->format->BytesPerPixel == 3) {		// slow 24 bpp mode
		Uint8 *bufp = (Uint8 *)img->pixels + y*img->pitch + x;
		bufp[img->format->Rshift/8] = c->r;
		bufp[img->format->Gshift/8] = c->g;
		bufp[img->format->Bshift/8] = c->b;
	} else {
		Uint32 mappedc = SDL_MapRGB(img->format, c->r, c->g, c->b);
		switch (img->format->BytesPerPixel) {
			case 1:		// indexed, palette
				((Uint8 *)img->pixels)[y*img->pitch + x] = mappedc;
				break;
			case 2:		// 15 or 16 bpp
				((Uint16 *)img->pixels)[y*img->pitch/2 + x] = mappedc;
				break;
			case 4:		// 32 bpp
				((Uint32 *)img->pixels)[y*img->pitch/4 + x] = mappedc;
				break;
		}
	}
}

static void sethorizline(SDL_Surface *img, SDL_Color *c, unsigned int x, unsigned int y, unsigned int w) {
	unsigned int i;
	if (img->format->BytesPerPixel == 3) {
		Uint8 *data = img->pixels;
		data += y*img->pitch + x;
		for (i=0; i<=w; i++) {
			(data + i)[img->format->Rshift/8] = c->r;
			(data + i)[img->format->Gshift/8] = c->g;
			(data + i)[img->format->Bshift/8] = c->b;
		}
	} else {
		Uint32 mappedc = SDL_MapRGB(img->format, c->r, c->g, c->b);
		switch (img->format->BytesPerPixel) {
			case 1: memset(img->pixels + y*img->pitch + x, mappedc, w+1); break;
			case 2: {
				Uint16 *data = img->pixels;
				data += y*img->pitch/2 + x;
				for (i=0; i<=w; i++)
					data[i] = mappedc;
				break;
			}
			case 4: {
				Uint32 *data = img->pixels;
				data += y*img->pitch/4 + x;
				for (i=0; i<=w; i++)
					data[i] = mappedc;
				break;
			}
		}
	}
}

static void setrect(SDL_Surface *img, SDL_Color *c, SDL_Color *fill, SDL_Rect *r) {
	int i = 0;
	if (fill) {
		Uint32 mc = SDL_MapRGB(img->format, fill->r, fill->g, fill->b);
		sdl_run(-1, FillRect, img, r, mc);
	}
	sethorizline(img, c, r->x       , r->y, r->w);
	while (++i < r->h) {
		setpixel(img, c, r->x       , r->y + i);
		setpixel(img, c, r->x + r->w, r->y + i);
	}
	sethorizline(img, c, r->x, r->y + r->h, r->w);
}

// for nicer colors see http://commons.wikimedia.org/w/index.php?title=File:Newton-lplane-Mandelbrot-smooth.jpg&oldid=15655904
static void get_color(SDL_Color *c, list_t *it) {
	unsigned int q, t, p = 0;
	unsigned int i, s, v;
	
	if (!it->root) {
		c->r = c->g = c->b = 0;
		return;
	}
	
	// represent the values in an HSV-color
	i = 0xFFFF/(it->itr+1);
	
	s = (i >> 8) & 0xFF;
	v = (i     ) & 0xFF;
	
	// calculate RGB-color from HSV
	t = (it->root->dhi.rem * s) / (6*HSV_ACC);
	q = v - t;
	switch (it->root->dhi.quot % 6) {
		case 0: c->r = v; c->g = t; c->b = p; break;
		case 1: c->r = q; c->g = v; c->b = p; break;
		case 2: c->r = p; c->g = v; c->b = t; break;
		case 3: c->r = p; c->g = q; c->b = v; break;
		case 4: c->r = t; c->g = p; c->b = v; break;
		case 5: c->r = v; c->g = p; c->b = q; break;
		default: return;
	}
}

static void render(SDL_Surface *img, render_t *rp) {
	root_t *nit;
	unsigned int x, y, i, j;
	
	if (SDL_MUSTLOCK(img))
		sdl_run(-1, LockSurface, img);
	
	i = 0;
	for (nit = rp->roots; nit != NULL; nit = nit->next) {
		unsigned int h = (i++ * 36 * HSV_ACC) / rp->nroots;
		nit->dhi = div(h, 6 * HSV_ACC);
	}
	
	for (x=0; x < rp->width; x += rp->step) {
		for (y=0; y < rp->height; y += rp->step) {
			SDL_Color c;
			get_color(&c, rp->litarr + y * rp->width + x);
			
			// paint the step x step pixels
			for (i=0; i<rp->step; i++)
				for (j=0; j<rp->step; j++)
					setpixel(img, &c, x + i, y + j);
		}
	}
	
	if (SDL_MUSTLOCK(img))
		SDL_UnlockSurface(img);
}

void sdl_gui(render_t *rp) {
	SDL_Surface *buffer;
	int flags;
	
	// init SDL
	flags = SDL_HWSURFACE | SDL_DOUBLEBUF | SDL_RESIZABLE;
	if (rp->fullscreen) flags |= SDL_FULLSCREEN;
// 	printf("XXX\n");
	buffer = sdl_init(flags, rp->width, rp->height);
	
	// render image
// 	printf("AAA\n");
	render(buffer, rp);
	// while not interrupted, waits for user-input and - if necessary - repaints the image
	mainloop(buffer, flags, rp);
	
	// cleanup
	SDL_FreeSurface(buffer);
	SDL_FreeSurface(display);
	SDL_Quit();
}

static SDL_Surface * sdl_init(int flags, int width, int height) {
	sdl_run(-1, Init, SDL_INIT_VIDEO | SDL_INIT_EVENTTHREAD);
	SDL_WM_SetCaption(CAPTION, TITLE);
	
	int rmask, gmask, bmask;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	rmask = 0xff000000;
	gmask = 0x00ff0000;
	bmask = 0x0000ff00;
#else
	rmask = 0x000000ff;
	gmask = 0x0000ff00;
	bmask = 0x00ff0000;
#endif
	SDL_Surface *buffer = SDL_CreateRGBSurface(flags & ~SDL_DOUBLEBUF, width, height, 32, rmask, gmask, bmask, 0);
	if (buffer == NULL)
		sdl_fail("SDL_CreateRGBSurface");
	
	if (false /* user did not set width and height explicitely */) {
		SDL_Rect **modes = SDL_ListModes(buffer->format, SDL_FULLSCREEN | flags);
		if (modes == NULL) {
			fprintf(stderr, "no video modes reported available, ignoring...\n");
		} else if (modes == (SDL_Rect **)-1) {
			printf("no restriction on video mode\n");
		} else {
			int i;
			for (i=0; modes[i]; i++)
				printf("  %dx%d\n", modes[i]->w, modes[i]->h);
		}
	}
	return buffer;
}

inline static SDL_Rect rect_from_abs(int x1, int x2, int y1, int y2) {
	SDL_Rect p;
	int d;
	if ((d = x1 - x2) < 0) { p.x = x1; p.w = -d; } else { p.x = x2; p.w = d; }
	if ((d = y1 - y2) < 0) { p.y = y1; p.h = -d; } else { p.y = y2; p.h = d; }
	return p;
}

static void mainloop(SDL_Surface * const buffer, int const flags, render_t *rp) {
	SDL_Surface *tmp = NULL;
	
	SDL_Color markc[] = {
			{ .r = 0xFF, .g = 0xFF, .b = 0xFF },		// white
			{ .r = 0x00, .g = 0x00, .b = 0x00 },		// black
	};
	
	/* needed for a filled selection rectangle to be painted
	SDL_Surface *overlay = SDL_CreateRGBSurface(SDL_HWSURFACE | SDL_SRCCOLORKEY | SDL_SRCALPHA, width, height, 8, 0, 0, 0, 0);
	sdl_run(0, SetColors, overlay, markc, 0, 2);
	sdl_run(-1, SetAlpha, overlay, SDL_SRCALPHA | SDL_RLEACCEL, 0x80);
	*/
	
	SDL_Rect marked = { 0, 0, 0, 0 };	// this rectangle 'marked' holds 2 (absolute) coords:
	SDL_Rect p = { 0, 0, 0, 0 }, prevp;	// (x,y): mouse was pressed and (w,h): current coords
	bool rerender = false;				// runs meshcalc_render for the new dimensions and iterations
	bool ismarked = false;				// if true: paints a rectangle around the marked area on the display
	bool markdone = false;				// todo: replace, when zooming is finished, then the last rect needn't be painted
	bool repaint;						// if true: blits the whole (possibly zoomed) buffer to the display
	
	while (!interrupted) {
		
		if (ismarked) {
			// calculate a rect with relative values w and h from the absolute coords in marked, as intended by SDL_Rect
			p = rect_from_abs(marked.x, marked.w, marked.y, marked.h);
		}
		
		if (rerender) {
			if (p.w != 0 && p.h != 0) {
				double complex low = rp->rng_lower, high = rp->rng_lower;
				low +=    (creal(rp->rng_upper - rp->rng_lower) / (rp->width  - 1) * (p.x      ));
				low +=  I*(cimag(rp->rng_upper - rp->rng_lower) / (rp->height - 1) * (p.y      ));
				high +=   (creal(rp->rng_upper - rp->rng_lower) / (rp->width  - 1) * (p.x + p.w));
				high += I*(cimag(rp->rng_upper - rp->rng_lower) / (rp->height - 1) * (p.y + p.h));
				
				printf("new range: [%g,%g] + i*[%g,%g]\n", creal(low), cimag(low), creal(high), cimag(high));
				// todo: set rng_lower and rng_upper to low and high
			}
			
			// run meshcalc_render for the new dimensions and itr and repaint
			
			rerender = false;
			repaint = true;
			markdone = true;
		}
		
		// width or height have changed, we need to resize display, possibly also the rendered image and blit it to display
		if (display == NULL || rp->width != (unsigned int)display->w || rp->height != (unsigned int)display->h) {
			display = SDL_SetVideoMode(rp->width, rp->height, 32, flags);
			if (display == NULL)
				sdl_fail("SDL_SetVideoMode");
			repaint = true;
		}
		
		if (repaint) {
			if (rp->width == (unsigned int)buffer->w && rp->height == (unsigned int)buffer->h) {
				if (tmp != NULL) {
					SDL_FreeSurface(tmp);
					tmp = NULL;
				}
			} else if (tmp == NULL || rp->width != (unsigned int)tmp->w || rp->height != (unsigned int)tmp->h) {
				if (tmp != NULL)
					SDL_FreeSurface(tmp);
				tmp = zoomSurface(buffer, (double)rp->width/buffer->w, (double)rp->height/buffer->h, SMOOTHING_ON);
			}
			sdl_run(-1, BlitSurface, (tmp == NULL) ? buffer : tmp, NULL, display, NULL);
			printf("repainted; width: %d, height: %d, rendered: [%d,%d]\n", rp->width, rp->height, buffer->w, buffer->h);
		}
		
		if (ismarked) {
			// todo: just blit the stuff that has changed
			if (!repaint)
				sdl_run(-1, BlitSurface, (tmp == NULL) ? buffer : tmp, &prevp, display, &prevp);
			
			if (p.w != 0 || p.h != 0) {
				if (SDL_MUSTLOCK(display)) sdl_run(-1, LockSurface, display);
				setrect(display, &markc[0], NULL, &p);
				if (SDL_MUSTLOCK(display)) SDL_UnlockSurface(display);
				p.w++;
				p.h++;
				// SDL_BlitSurface(overlay, &p, display, &p);
			}
			prevp = p;
		}
		
		// update the display
		SDL_Flip(display);
		
		repaint = false;
		p.x = p.y = p.w = p.h = 0;
			
		if (markdone) {
			markdone = false;
			ismarked = false;
			marked.x = marked.y = marked.w = marked.h = 0;
		}
		
		
		bool refresh = false;		// if set to true, means that the queue is purged and then the display refreshed
		
		while (!interrupted) {
			SDL_Event event;
			int ret = (refresh) ? SDL_PollEvent(&event) : SDL_WaitEvent(&event);
			if (!ret) break;
			
			switch (event.type) {
			case SDL_KEYDOWN:
				switch (event.key.keysym.sym) {
				case SDLK_q:
				case SDLK_ESCAPE:
					goto end;
				case SDLK_f:
					sdl_run(1, WM_ToggleFullScreen, display);
					refresh = true;
					break;
				case SDLK_w: {
					event.type = SDL_VIDEORESIZE;
					event.resize.w = buffer->w;
					event.resize.h = buffer->h;
					SDL_PushEvent(&event);
					break;
				}
				case SDLK_s: {
					size_t flen = strlen(rp->f.fstr);
					flen += 1 + 12 + strlen("save__.bmp");
					char *buf = malloc(flen);
					snprintf(buf, flen, "save_%s_%lu.bmp", rp->f.fstr, time(NULL));
					SDL_SaveBMP(buffer, buf);
					printf("saved image to '%s'\n", buf);
					free(buf);
					break;
				}
				
				case SDLK_PAGEUP:
					rp->itr = min(rp->itr * 2, MAX_ITR);
					printf("increased number of iterations to %d\n", rp->itr);
					rerender = true;
					refresh = true;
					break;
				case SDLK_PAGEDOWN:
					rp->itr = max(rp->itr / 2, MIN_ITR);
					printf("decreased number of iterations to %d\n", rp->itr);
					rerender = true;
					refresh = true;
					break;
				
				case SDLK_PLUS:
					p.x = (rp->width-1) / 4;
					p.y = (rp->height-1) / 4;
					p.w = rp->width / 2;
					p.h = rp->height / 2;
					refresh = true;
					rerender = true;
					break;
				case SDLK_MINUS:
					p.x = -(rp->width-1)/2;
					p.y = -(rp->height-1)/2;
					p.w = 2*rp->width-1;
					p.h = 2*rp->height-1;
					refresh = true;
					markdone = true;
					rerender = true;
					break;
				
				case SDLK_UP:
				case SDLK_DOWN:
				case SDLK_LEFT:
				case SDLK_RIGHT:
					p.x = 0;
					p.y = 0;
					p.w = rp->width;
					p.h = rp->height;
					switch (event.key.keysym.sym) {
					case SDLK_UP:    p.y -= rp->height / 3; break;
					case SDLK_DOWN:  p.y += rp->height / 3; break;
					case SDLK_LEFT:  p.x -= rp->width  / 3; break;
					case SDLK_RIGHT: p.x += rp->width  / 3; break;
					default:;
					}
					refresh = true;
					markdone = true;
					rerender = true;
					break;
				case SDLK_h:
					printf( "q / ESC   : quit\n"
							"f         : toggle fullscreen mode\n"
							"w         : restore window size to rendered size\n"
							"s         : save rendered image to file\n"
							"pg up/down: in-/decrease number of iterations\n"			// need list of black pixels
							"+ / -     : zoom in/out\n"
							"arrows    : move displayed section up/down/left/right\n"	// need to specify render area
							"h         : print this help message\n");
					break;
				default:;
				}
				break;
			
			case SDL_MOUSEMOTION:
				if (event.motion.state & SDL_BUTTON(SDL_BUTTON_LEFT)) {
					marked.w = event.motion.x;
					marked.h = event.motion.y;
					refresh = true;
				}
				break;
			case SDL_MOUSEBUTTONUP:
				if (event.button.button == SDL_BUTTON_LEFT) {
					marked.w = event.button.x;
					marked.h = event.button.y;
					refresh = true;
					markdone = true;
					rerender = true;
				} else if (event.button.button == SDL_BUTTON_RIGHT) {
					p.x = -(rp->width-1)/2;
					p.y = -(rp->height-1)/2;
					p.w = 2*rp->width-1;
					p.h = 2*rp->height-1;
					refresh = true;
					markdone = true;
					rerender = true;
				}
				break;
			
			case SDL_MOUSEBUTTONDOWN:
				if (event.button.button == SDL_BUTTON_LEFT) {
					marked.x = event.button.x;
					marked.y = event.button.y;
					ismarked = true;
				}
				break;
			
			case SDL_QUIT:
				goto end;
			
			case SDL_VIDEORESIZE:
				rp->width = event.resize.w;
				rp->height = event.resize.h;
				// fall through
			case SDL_VIDEOEXPOSE:
				refresh = true;
				break;
			}
		}
	}

end:
	if (tmp != NULL) SDL_FreeSurface(tmp);
}
