
/* run with i.e.:
 * - newton -f 'z*(z*(z+1)-2)-1' 'z*(z*3+2)-2' -v -40,30 -50,50 700 1000
 * - newton -f 'z*z*z-1' '3*z*z' -v 1000
 * - newton -f 'z*z*z*(1+z^12*(1+z^30))+1+exp(1/z)' 'z*z*(3+z^12*(15+45*z^30))+log(z)*exp(1/z)' -e 1/10^10 -i 800 -.8,.8 1000 -v
 */

#define _XOPEN_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <math.h>
#include <libgen.h>		/* for basename() */
#include <unistd.h>		/* for pause() */
#include <stdbool.h>
#include <complex.h>
#include <string.h>

#include <time.h>
#include <sys/time.h>

#define tdelta_us(a, b) (((b).tv_sec - (a).tv_sec) * 1000000 + (b).tv_usec - (a).tv_usec)
#define tdelta_ms(a, b) (((b).tv_sec - (a).tv_sec) * 1000 + ((b).tv_usec - (a).tv_usec) / 1000)

#include "formparse.h"
#include "eval.h"
#include "cgamma.h"

#define NEXTERN
#include "newton.h"
#undef NEXTERN

#define DEFAULT_WIDTH 400
#define DEFAULT_HEIGHT DEFAULT_WIDTH
#define DEFAULT_RANGE_LOWER (-1.0 + I * -1.0)
#define DEFAULT_RANGE_UPPER (-DEFAULT_RANGE_LOWER)
#define DEFAULT_EPS "10^(-14)"
#define DEFAULT_ITR 200

static double complex _cabs(double complex a)   { return (double complex)cabs(a); }
static double complex _creal(double complex z)  { return (double complex)creal(z); }
static double complex _cimag(double complex z)  { return (double complex)cimag(z); }
static double complex _carg(double complex z)   { return (double complex)carg(z); }
static double complex _cgamma(double complex z) { return cgamma(z, NULL); }

// todo: how to distinguish between using the real or the complex version of the function? abandon the reals?
static mfuncs_t mfunctions[] = {
		{ NULL,   conj,    "conj",   "Complex conjugate" },
		{ NULL,   _creal,  "re",     "Real part of the complex value" },
		{ NULL,   _cimag,  "im",     "Imaginary part of the complex value" },
		{ fabs,   _cabs,   "abs",    "Absolute value / Norm / Modulus" },
		{ NULL,   _carg,   "arg",    "Argument / Angle of the complex value" },
		{ sin,    csin,    "sin",    "Sine" },
		{ asin,   casin,   "arcsin", "Arcsine" },
		{ cos,    ccos,    "cos",    "Cosine" },
		{ acos,   cacos,   "arccos", "Arccosine" },
		{ tan,    ctan,    "tan",    "Tangent" },
		{ atan,   catan,   "arctan", "Arctangent" },
		{ exp,    cexp,    "exp",    "Exponential function" },
		{ log,    clog,    "log",    "Natural logarithm" },
		{ sqrt,   csqrt,   "sqrt",   "Square root" },
		{ tgamma, _cgamma, "gamma",  "Factorial / Gamma function" },
		{ lgamma, lcgamma, "lgamma", "Logarithm of the gamma function" },
		{ NULL, NULL, NULL, NULL }
};
static mvars_t mvariables[] = {
		{ false, 0,    "x",  "Real part of complex variable z" },
		{ false, 0,    "y",  "Imaginary part of complex variable z" },
		{ false, 0,    "z",  "Complex variable z = x + y*i" },
		{ true,  M_PI, "pi", "Pi" },
		{ true,  M_E,  "e",  "Euler's number e" },
		{ true,  I,    "i",  "Imaginary unit i" },
		{ false, 0, NULL, NULL }
};

typedef struct {
	double complex arg;
	double complex fval;
	const double eps;					// todo: make const
	root_t *nl;
	form_t *f, *df;
	const unsigned int maxitr;
	unsigned int itr;
	unsigned int nsize;
	unsigned int mitr;
} mesharg_t;

static char *program_name;
static bool benchmark = false;
bool interrupted = 0;							// set to true when the user interrupts via ^C or so, todo: move to where guis may access the variable

#ifdef STATS
static unsigned int *itrs;
#endif

static double complex __fastcall newton(mesharg_t * const st) {
	unsigned int i;
	const unsigned int mitr = st->maxitr;
	double complex fz = NAN + I*NAN, z = st->arg;
	double complex var[3];
	
	// use fixed points for iterations, if after x itrs the error is small enough, quit
	for (i=0; i<mitr; i++) {
		var[0] = creal(z);
		var[1] = cimag(z);
		var[2] = z;
		fz = calc(st->f, var);				// todo: this returns just one skalar, so |var| should be 1 (?)
		if (cabs(fz) <= st->eps)
			break;
		double complex dfz = calc(st->df, var);
		z -= fz/dfz;
	}
	
	st->itr = i;
	st->fval = fz;
	return z;
}

static int __fastcall meshexec(list_t * const it, mesharg_t * const marg) {
	root_t *nit;
	int test;
	double complex res;
	double eps;
	
	if (interrupted)
		return -1;
	
	res = newton(marg);
	
#ifdef STATS
	itrs[marg->itr]++;
#endif
	
	// check, whether the result makes sense
	test = fpclassify(res);
	if (test == FP_NAN || test == FP_INFINITE)
		return 1;
	if (marg->itr > marg->mitr)
		marg->mitr = marg->itr;
	if (cabs(marg->fval) > marg->eps)
		return 1;
	
	it->z = res;
	it->fz = marg->fval;
	it->itr = marg->itr;
	
	eps = marg->eps * 2;
	for (nit = marg->nl; nit != NULL; nit = nit->next)		// xxx: maybe use tree: O(n) -> O(ld n)
		if (cabs(nit->z - res) <= eps)
			break;
	
	// not found -> insert
	if (nit == NULL) {
		nit = malloc(sizeof(root_t));
		nit->z = res;
		nit->next = marg->nl;
		marg->nl = nit;
		marg->nsize++;
	}
	it->root = nit;
	return 0;
}

static void meshcalcroots(render_t *rp) {
	static char const *msg = "calculating line %d/%d, roots found: %d\r";
	struct timeval s, t;
	double complex yarg;
	unsigned int x, y;
	list_t *litarr = rp->litarr;
	mesharg_t marg = {
		.nl = NULL,
		.nsize = 0,
		.eps = rp->eps / 2,
		.mitr = 0,
		.f = &rp->f,
		.df = &rp->df,
		.maxitr = rp->itr
	};
	
	gettimeofday(&s, NULL);
	if (rp->maprngx == NULL && rp->maprngy == NULL) {
		// linear mapping of the pixels to the given range
		double const xoff = creal(rp->rng_lower);
		double const yoff = cimag(rp->rng_lower);
		double const xrange = creal(rp->rng_upper - rp->rng_lower)/(rp->width-1);
		double const yrange = cimag(rp->rng_upper - rp->rng_lower)/(rp->height-1);
		for (y = rp->off; y < rp->height && !interrupted; y += rp->step) {
			if (rp->verbose)
				printf(msg, y, rp->height, marg.nsize);
			fflush(stdout);
			yarg = xoff + I*yoff + I*(y * yrange);
			for (x = rp->off; x < rp->width; x += rp->step) {
				marg.arg = yarg + (x*xrange);
				if (meshexec(litarr + y * rp->width + x, &marg) == -1)
					break;
			}
		}
	} else {
		// custom mapping of the pixels to the given range, defined by maprngy and maprngx
		for (y = rp->off; y < rp->height && !interrupted; y += rp->step) {
			if (rp->verbose)
				printf(msg, y, rp->height, marg.nsize);
			fflush(stdout);
			yarg = I*rp->maprngy(y);
			for (x = rp->off; x < rp->width; x += rp->step) {
				marg.arg = rp->maprngx(x) + yarg;
				if (meshexec(litarr + y * rp->width + x, &marg) == -1)
					break;
			}
		}
	}
	gettimeofday(&t, NULL);
	
	printf("found %d roots of '%s' for %d*%d points with max. %d iterations in %lu ms (%.0f points/sec)\n",
			marg.nsize, rp->f.fstr, 1 + (rp->width-1)/rp->step, y/rp->step,
			marg.mitr, tdelta_ms(s, t), ((1 + (y * (rp->width-1)) / rp->step) * 1000.0) / tdelta_ms(s, t));
	
	rp->roots = marg.nl;
	rp->nroots = marg.nsize;
}

	/* todo
	 * V 1  create grammar
	 * V 2  parse formula into f (needs: lexer (whitespace, etc.), parser)
	 * x 3  create differential f'
	 * v 4  optimize f and f': run calc() but don't resolve variables => const. exprs get pre-calculated
	 * - 5  determine nst of the function
	 *
	 * and/or: instead of 3 and 5 let the user supply a formula-config-file, which additionally to the
	 *         above information includes the nst and the colors for them
	 *
	 * - user may specify a function rangemap : [0,width)x[0,height) -> [rng_lower,rng_upper] instead of the range
	 *
	 * formparse: remove exit()-calls, return NULL instead and provide a ptree_strerr()-function
	 *            add support for factor0s like 1e-42
	 *            add support for terms like 1*-1 and 1^-1 (without the parantheses around the 2. argument of the op)
	 * eval: need to implement associativity and commutativity to be able to minimize more efficiently:
	 *         - assoc: tree-rotations
	 *       add support for replacement of x^0.5 by sqrt(x), e^x by exp(x) and the like (calc faster and less states)
	 *       remove operations including the one-element 0 from + as well as 0 and 1 from * and ^
	 */
#define min(a, b) (((a) < (b)) ? (a) : (b))
#define max(a, b) (((a) > (b)) ? (a) : (b))

static void meshcalc(render_t *rp) {
// 	root_t *nit;
// 	list_t *it;
	
	unsigned int size = rp->width * rp->height * sizeof(list_t);
	rp->litarr = realloc(rp->litarr, size);
	memset(rp->litarr, 0, size);
	
#ifdef STATS
	itrs = calloc(rp->itr + 1, sizeof(unsigned int));
#endif
	
	// calculate the values for each pixel
	meshcalcroots(rp);
	
#ifdef STATS
	unsigned int i, wacc = 0, acc = 0, max = 0;
	for (i=0; i<=rp->itr; i++)
		if (itrs[i] > 0) {
			wacc += i * itrs[i];
			acc += itrs[i];
			if (itrs[i] > itrs[max])
				max = i;
			if (rp->verbose > 1)
				printf("%3d: %u\n", i, itrs[i]);
		}
	free(itrs);
	printf("number of iterations: avg: %.1f, principal: %u, \n", wacc / (float)acc, max);
#endif
	
	// todo: use arrays of dim nroots
	/*
	for (nit = rp->roots; nit != NULL; nit = nit->next) {
		int lsize = 0;
		double sumabs = 0;
		double complex sumz = 0;
		for (it = nit->val; it != NULL; it = it->next) {
			sumabs += cabs(it->fz);
			sumz += it->z;
			lsize++;
		}
		nit->z = sumz / lsize;
		nit->accuracy = sumabs / lsize;
		if (rp->verbose > 1)
			printf("\rroot @ z = %g + %gi\t(|f(z)| = %g)\n", creal(nit->z), cimag(nit->z), nit->accuracy);
	}*/
}

static void usage(int);
static void parse_args(int, char **, render_t *);
static void handler(int);

int main(int argc, char **argv) {
	render_t rp = {
		.step = 1,
		.off = 0,
		.width = DEFAULT_WIDTH,
		.height = DEFAULT_HEIGHT,
		.itr = DEFAULT_ITR,
		.rng_lower = DEFAULT_RANGE_LOWER,
		.rng_upper = DEFAULT_RANGE_UPPER,
		.verbose = false,
		.fullscreen = false,
		.maprngx = NULL,
		.maprngy = NULL,
		.litarr = NULL,
	};
	root_t *nit;
	
	program_name = basename(*argv++);
	parse_args(argc - 1, argv, &rp);
	
	// register signal handler
	signal(SIGINT, handler);
	signal(SIGTERM, handler);
	
	// run newton for each value corresponding to a pixel and render the pixel
	meshcalc(&rp);
	
	// todo: test hsv-generation for n = 15 and n = 45 (and possibly higher values for n)
	if (!benchmark) {
#ifndef NO_SDL
		sdl_gui(&rp);
#endif
	}
	
	while ((nit = rp.roots) != NULL) {
		rp.roots = nit->next;
		free(nit);
	}
	free(rp.litarr);
	
	form_free(&rp.f);
	form_free(&rp.df);
	
	exit(EXIT_SUCCESS);
}

static void handler(int sig) {
	switch (sig) {
		case SIGINT:
		case SIGTERM:
			interrupted = true;
			break;
	}
	signal(sig, handler);
}

static void post_prep_form(bool verbose, form_t *f, char *label) {
	if (verbose) {
		printf("parsed (%d nodes): %s = ", f->node_cnt, label);
		print_stack(f->storg, f->node_cnt, mfunctions);
	}
	minimize(f);
	if (verbose) {
		printf("; minimized (%d nodes): %s = ", f->node_cnt, label);
		print_stack(f->storg, f->node_cnt, mfunctions);
		printf("\n");
	}
}

#define err0(fmt, ...) do { \
			fprintf(stderr, "Illegal argument: " fmt "\n", ##__VA_ARGS__); \
			exit(EXIT_FAILURE); \
		} while (0)
#define err1() err0("%s", argv[i])

static void parse_args(int argc, char **argv, render_t *rp) {
	int i, j;
	char *endptr, *sep, *eps_str = DEFAULT_EPS;
	enum { A, B, C, D } state = A;
	long l;
	ptree_t *t;
	form_t feps;
	
	int vrsc = sizeof(mvariables)/sizeof(mvariables[0]) - 1;
	int fncc = sizeof(mfunctions)/sizeof(mfunctions[0]) - 1;
	const char *vrsv[vrsc], *fncv[fncc];
	for (j=0; j<vrsc; j++) vrsv[j] = mvariables[j].name;
	for (j=0; j<fncc; j++) fncv[j] = mfunctions[j].name;
	
	
	for (i=0; i<argc; i++) {
		// help message
		if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help"))
			usage(EXIT_SUCCESS);
		
		// formulas
		if (!strcmp(argv[i], "-f")) {
			if (i + 2 >= argc) 
				err0("'-f': both formulas f(x) and f'(x) must be provided");
			char *label[] = { "f(z) ", "f'(z)" };
			form_t *forms[] = { &rp->f, &rp->df };
			
			// parse f(x) and f'(x) and init the respective stacks and output parsed formula
			form_out_infix = (rp->verbose) ? 1 : 0;
			for (j=0; j<2; j++, i++) {
				t = parse(argv[i+1], vrsc, vrsv, fncc, fncv);
				if (t == NULL) err0("error parsing formula %s = '%s'", label[j], argv[i+1]);
				if (rp->verbose) printf("%s: ", label[j]);
				form_prep(argv[i+1], t, forms[j], mvariables, mfunctions, 0);
				if (rp->verbose) printf("\t%d nodes\n", forms[j]->node_cnt);
				// form_prep() has already taken care of free'ing the ptree_t-structures
			}
			form_out_infix = 0;
			
			continue;
		}
		
		// epsilon
		if (!strcmp(argv[i], "-e")) {
			if (++i == argc)
				err0("'-e' needs an argument");
			eps_str = argv[i];
			continue;
		}
		
		if (!strcmp(argv[i], "-v")) {
			rp->verbose++;
			continue;
		}
		
		if (!strcmp(argv[i], "-F")) {
			rp->fullscreen = true;
			continue;
		}
		
#define read0(fun, check, v) do { \
			v = fun; \
			if (errno) err0("%s: %s", argv[i], strerror(errno)); \
			if (*endptr != '\0' || check) err1(); \
		} while (0)
		
		if (!strcmp(argv[i], "-i")) {
			if (++i == argc)
				err0("'-i' needs an argument");
			long int a;
			read0(strtol(argv[i], &endptr, 0), a < MIN_ITR || a > MAX_ITR, a);
			rp->itr = a;
			continue;
		}
		
		if (!strcmp(argv[i], "-b")) {
			benchmark = true;
			continue;
		}
		
		if (state > D)
			err0("reached illegal state %d: too many arguments", state);
		
		if ((sep = strchr(argv[i], ',')) != NULL) {
			if (state > B) err1();
			
			*sep = '\0';
			
			double xl, xh;
			read0(strtod(argv[i], &endptr), 0, xl);
			read0(strtod(sep + 1, &endptr), 0, xh);
			
			if (state == A) {
				rp->rng_lower = xl + I*xl;
				rp->rng_upper = xh + I*xh;
			} else {
				rp->rng_lower = creal(rp->rng_lower) + I*xl;
				rp->rng_upper = creal(rp->rng_upper) + I*xh;
			}
		} else {
			if (state < C) state = C;
			
			read0(strtol(argv[i], &endptr, 0), l <= 1, l);
			*((state == C) ? &rp->width : &rp->height) = l;
		}
#undef read0
		
		state++;
	}
	
#define cratio(z1, z2) ((creal(z2)-creal(z1))/(cimag(z2)-cimag(z1)))		/* w/h */
	
	if (state <= D)		// user hasn't entered height explicitely, so set it according to the range's aspect ratio
		rp->height = rint(rp->width / cratio(rp->rng_lower, rp->rng_upper));
	
	if (creal(rp->rng_lower) > creal(rp->rng_upper)) err0("xrange illegal: x1 < x2, [%g,%g]", creal(rp->rng_lower), creal(rp->rng_upper));
	if (cimag(rp->rng_lower) > cimag(rp->rng_upper)) err0("yrange illegal: y1 < y2, [%g,%g]", cimag(rp->rng_lower), cimag(rp->rng_upper));
	
	if (rp->f.storg == NULL || rp->df.storg == NULL) err0("both formulas f(x) and f'(x) must be provided");
	
	// parse epsilon argument
	t = parse(eps_str, vrsc, vrsv, fncc, fncv);
	if (t == NULL)
		err0("error parsing eps-formula '%s'", eps_str);
	form_prep(eps_str, t, &feps, mvariables, mfunctions, 0);
	minimize(&feps);
	if (feps.node_cnt > 1)
		err0("'-e': '%s' is not a valid constant expression, has %d nodes", eps_str, feps.node_cnt);
	if (cimag(feps.storg->data.val) != 0.0)
		err0("'-e': %s = %g + %gi contains an imaginary part, which must be zero for the euklidean distance",
				eps_str, creal(feps.storg->data.val), cimag(feps.storg->data.val));
	if (creal(feps.storg->data.val) < 0.0)
		err0("'-e': %s = %g is negative, but the euklidean distance must be non-negative",
				eps_str, creal(feps.storg->data.val));
	rp->eps = feps.storg->data.val;
	
	// some stats
	printf("plane dimensions: %ux%u px, plot range: [%g,%g] x i*[%g,%g], aspect: %g:%g, epsilon: %g\n",
			rp->width, rp->height,
			creal(rp->rng_lower), creal(rp->rng_upper), cimag(rp->rng_lower), cimag(rp->rng_upper),
			(double)rp->width / (double)rp->height, cratio(rp->rng_lower, rp->rng_upper), rp->eps);
	
	post_prep_form(rp->verbose, &rp->f, "f(z)");
	post_prep_form(rp->verbose, &rp->df, "f'(z)");
}

#undef cratio
#undef err1
#undef err0

static void usage(int ecode) {
	unsigned int i;
	fprintf(stderr, "Usage: %s [-v] [-F] [-f FORMULA DIFFERENTIAL] [-e EPS]" /* " [-t THREADS]" */ "\n", program_name);
	for (i=0; i<strlen(program_name); i++) fprintf(stderr, " ");
	fprintf(stderr, "        [XLOW,XHIGH [YLOW,YHIGH]] [WIDTH [HEIGHT]]\n");
	fprintf(stderr, "\n"
					"The formula may contain the usual operators: +,-,*,/,^,! and groupings in\n"
					"parantheses. The variables are interpreted in the complex plane, whereas\n"
					"z = x + i*y; others than x,y,z are not allowed.\n");
	
	mvars_t *mv = mvariables;
	if (mv->name != NULL) {
		fprintf(stderr, "Additionally the following constants may be used:\n");
		for (mv = mvariables; mv->name != NULL; mv++)
			if (mv->isconstant)
				fprintf(stderr, "  %-2s = %7g+%-1gi  %s\n", mv->name, creal(mv->val), cimag(mv->val), mv->desc);
	}
	
	mfuncs_t *mf = mfunctions;
	if (mf->name != NULL) {
#define supp(w) ((mf->w ## fun == NULL) ? " " : #w)
		fprintf(stderr, "The following functions are supported (r: real, c: complex):\n");
		for (mf = mfunctions; mf->name != NULL; mf++)
			fprintf(stderr, "  [%s%s] %-8s %s\n", supp(r), supp(c), mf->name, mf->desc);
#undef supp
	}
#define str(x) #x
#define xstr(x) str(x)
	fprintf(stderr, "\n"
					"Arguments:\n"
					"  -f FORMULA   sets the function f(z) for which the Newton-Iterations will be\n"
					"               calculated. Remember to escape any symbols that may be interpre-\n"
					"               ted by your shell.\n"
					"  -e EPS       maximal difference between two calculated values to consider them\n"
					"               equal. format is a formula without any vars, default is " DEFAULT_EPS "\n"
					"  -i ITR       maximal number of iterations to perform, default is " xstr(DEFAULT_ITR) "\n"
//					"  -t THREADS   number of concurrent threads to use for rendering\n"
#ifndef NO_SDL
					"  -F           fullscreen mode (toggle via 'f' when the application runs)\n"
					"  -b           enable benchmark mode: no graphics are generated, just the plain\n"
					"               calculations are performed\n"
#endif
					"  -h           displays this help message\n"
					"\n"
					"This program is Free Software and released under the terms of the General Public\n"
					"License v2 by Franz Brausse (<dev@karlchenofhell.org>) in the hope to be useful.\n");
#undef str
#undef xstr
	exit(ecode);
}
